@@ -15,6 +15,25 @@ objects from a Tangram-compatible object specification.
 
 =head1 SYNOPSIS
 
+ # simple package - four attributes "foo", "bar" (integers),
+ # "baz" and "quux" (strings)
+ package MyObject;
+
+ use base qw(Class::Tangram);
+
+ our $schema = {
+    fields => { int => [ qw(foo bar) ],
+                string => [ qw(baz quux) ] }
+    };
+
+ package main;
+
+ my $object = MyObject->new(foo => 2, baz => "hello");
+ print $object->foo();
+ $object->set_quux("Something");
+
+ # More detailed example, displaying many of the features of
+ # Class::Tangram
  package Orange;
 
  use base qw(Class::Tangram);
@@ -46,36 +65,53 @@ objects from a Tangram-compatible object specification.
          ref => {
              grower => undef,
          },
+
+         # 'required' attributes - insist that these fields are set,
+         #    both with constructor and set()/set_X methods
+	 string => {
+             # true: 'type' must have non-empty value
+	     type => { required => 1 },
+	     # false: 'tag' must be defined but may be empty
+	     tag => { required => '' },
+         },
          # fields allowed by Class::Tangram but not ever
-         # stored - no type checking
+         # stored - no type checking without check_func
          transient => [ qw(_tangible) ],
      },
  };
  Class::Tangram::import_schema("Orange");
 
  package Project;
- # here's where we build the individual object schemas into a
- # Tangram::Schema object, which the Tangram::Storage class uses to
- # know which tables and columns to find objects.
+
+ # here's where we build the individual object schemas into
+ # a Tangram::Schema object, which the Tangram::Storage
+ # class uses to know which tables and columns to find
+ # objects.
  use Tangram::Schema;
 
  my $dbschema = Tangram::Schema->new
-     ({ classes => [ 'Orange' => $Orange::schema ]});
+     ({ classes => [ 'Orange' => $Orange::schema,
+                     'MyObject' => $MyObject::schema ]});
 
  sub schema { $dbschema };
 
  package main;
 
- # See Tangram::Relational for instructions on using "deploy" to
- # create the database this connects to.  You only have to do this if
- # you want to write the objects to a database.
+ # See Tangram::Relational for instructions on using
+ # "deploy" to create the database this connects to.  You
+ # only have to do this if you want to write the objects to
+ # a database.
  use Tangram::Relational;
  my ($dsn, $u, $p);
- my $storage = Tangram::Relational->connect(Project->schema,
-                                            $dsn, $u, $p);
+ my $storage = Tangram::Relational->connect
+                   (Project->schema, $dsn, $u, $p);
 
  # This is how you create instances
- my $orange = Orange->new(juiciness => 8);
+ my $orange = Orange->new(
+			  juiciness => 8,
+			  type => 'Florida',
+			  tag => '',  # required
+			 );
 
  # Store them
  $storage->insert($orange);
@@ -88,9 +124,13 @@ objects from a Tangram-compatible object specification.
  $orange->set_grower ($grower);
 
  # these are all illegal - type checking is fairly strict
+ my $orange = eval { Orange->new; };         print $@;
  eval { $orange->set_juiciness ("Yum"); };   print $@;
  eval { $orange->set_segments (31); };       print $@;
  eval { $orange->set_grower ("Mr. Nice"); }; print $@;
+ eval { $orange->set_type (''); };           print $@;
+ eval { $orange->set_type (undef); };        print $@;
+ eval { $orange->set_tag (undef); };         print $@;
 
  # if you prefer
  $orange->get( "juiciness" );
@@ -111,7 +151,8 @@ an exported variable C<$schema>.
 After writing this I found that it was useful for merely adding type
 checking and validation to arbitrary objects.  There are several
 modules on CPAN to do that already, but many don't have finely grained
-type checking, and none of them integrated with Tangram.
+type checking, and none of them integrated with Tangram or any other
+object persistence framework quite so easily.
 
 =head1 DEPENDENCIES
 
@@ -130,7 +171,7 @@ of the above modules, please send me an e-mail.
 
 =head2 MODULE RELEASE
 
-This is Class::Tangram version 1.09.
+This is Class::Tangram version 1.10.
 
 =cut
 
@@ -142,7 +183,7 @@ use vars qw($AUTOLOAD $VERSION %defaults);
 use Set::Object;
 use Pod::Constants -trim => 1,
     'MODULE RELEASE' => sub { ($VERSION) = /(\d+\.\d+)/ },
-    'TANGRAM TYPE TO FUNCTION MAPPING' => sub { %defaults = eval };
+    'Default Type Checking' => sub { %defaults = eval };
 
 local $AUTOLOAD;
 
@@ -166,17 +207,23 @@ my (%cleaners);
 # $init_defaults{$class}->{$attribute}
 my (%init_defaults);
 
+# $required_attributes{$class}->{$attribute} records which attributes
+# are required... used only by new() at present.
+my (%required_attributes);
+
 # if a class is abstract, complain if one is constructed.
 my (%abstract);
 
 =head1 METHODS
 
+=head2 Constructor
+
 =over 4
 
 =item Class-E<gt>new (attribute1 =E<gt> value, attribute2 =E<gt> value)
 
-sets up a new object of type Class, with attributes set to the values
-supplied.
+sets up a new object of type C<Class>, with attributes set to the
+values supplied.
 
 Can also be used as an object method, in which case it returns a
 B<copy> of the object, without any deep copying.
@@ -211,38 +258,52 @@ sub new ($@)
 	$self->set (@values); # start with @values
 
 	# now fill in fields that have defaults
-	for my $attribute (keys %{$init_defaults{$class}}) {
+	while ( my ($attribute, $default) =
+		each %{$init_defaults{$class}} ) {
 
 	    next if (exists $self->{$attribute});
 
-	    my $default = $init_defaults{$class}->{$attribute}
-		unless tied $init_defaults{$class}->{$attribute};
-
 	    if (ref $default eq "CODE") {
 		# sub { }, attribute gets return value
-		$self->{$attribute}
-		    = $init_defaults{$class}->{$attribute}->($self);
+		$self->{$attribute} = $default->($self);
 
 	    } elsif (ref $default eq "HASH") {
 		# hash ref, copy hash
-		$self->{$attribute}
-		    = { %{ $init_defaults{$class}->{$attribute} } };
+		$self->{$attribute} = { %{ $default } };
 
 	    } elsif (ref $default eq "ARRAY") {
 		# array ref, copy array
-		$self->{$attribute}
-		    = [ @{ $init_defaults{$class}->{$attribute} } ];
+		$self->{$attribute} = [ @{ $default } ];
 
 	    } else {
 		# something else, an object or a scalar
-		$self->{$attribute}
-		    = $init_defaults{$class}->{$attribute};
+		$self->{$attribute} = $default;
 	    }
 	}
+
+        # make sure field is not undef if 'required' option is set
+	#     ** UNLESS ** we're loading from the database
+	if (my $required = $required_attributes{$class}
+	        and caller !~ /^Tangram::/) {
+
+	    my @missing;
+	    while ( my ($attribute, $value) = each %$required ) {
+		push(@missing, $attribute)
+		    if ! exists $self->{$attribute};
+	    }
+	    croak("constructor call missing required attribute(s): "
+	          .join(', ',@missing).'.') if @missing;
+	}
     }
     return $self;
 }
 
+=back
+
+=head2 Accessing & Setting Attributes
+
+=over
+
 =item $instance->set(attribute => $value, ...)
 
 Sets the attributes of the given instance to the given values.  croaks
@@ -278,10 +339,10 @@ sub set($@) {
 
 =item $instance->get($attribute)
 
-Gets the value of $attribute.  If the attribute in question is a set,
-and this method is called in list context, then it returns the MEMBERS
-of the set (if called in scalar context, it returns the Set::Object
-container).
+Gets the value of C<$attribute>.  If the attribute in question is a
+set, and this method is called in list context, then it returns the
+MEMBERS of the set (if called in scalar context, it returns the
+Set::Object container).
 
 =cut
 
@@ -319,20 +380,20 @@ sub get($$) {
 
 =item $instance->attribute($value)
 
-If $value is not given, then
-this is equivalent to $instance->get("attribute")
+If C<$value> is not given, then
+this is equivalent to C<$instance-E<gt>get("attribute")>
 
-If $value is given, then this is equivalent to
-$instance->set("attribute", $value).  This usage issues a warning; you
-should change your code to use the set_attribute syntax for better
-readability.
+If C<$value> is given, then this is equivalent to
+C<$instance-E<gt>set("attribute", $value)>.  This usage issues a
+warning; you should change your code to use the set_attribute syntax
+for better readability (object methods should always be a verb).
 
 =item $instance->get_attribute
 
 =item $instance->set_attribute($value)
 
-Equivalent to $instance->get("attribute") and $instance->set(attribute
-=> $value), respectively.
+Equivalent to C<$instance-E<gt>get("attribute")> and
+C<$instance-E<gt>set(attribute =E<gt> $value)>, respectively.
 
 =item $instance->attribute_includes(@objects)
 
@@ -344,8 +405,9 @@ Equivalent to $instance->get("attribute") and $instance->set(attribute
 
 =item $instance->attribute_remove(@objects)
 
-Equivalent to calling $instance->attribute->includes(@objects), etc.
-This only works if the attribute in question is a Set::Object.
+Equivalent to calling
+C<$instance-E<gt>attribute-E<gt>includes(@objects)>, etc.  This only
+works if the attribute in question is a "set" or "iset" type.
 
 =cut
 
@@ -414,43 +476,88 @@ sub getset($$;$) {
 
 }
 
+=back
+
+=head2 Attribute Type Checking
+
+Class::Tangram provides type checking of attributes, which are defined
+via three per-attribute options:
+
+=over
+
+=item check_func
+
+A function that is called with a reference to the new value in
+C<$_[0]>.  It should call C<die()> if the value is bad.
+
+=item destroy_func
+
+If anything special needs to happen to this attribute before the
+object is destroyed (or when someone calls
+C<$object-E<gt>clear_refs()>), then define this.
+
+=item required
+
+If this option is set to a true value, then the attribute must be set
+to a true value to pass type checking.  If it is defined but logically
+false, (ie "" or 0), then the attribute must also be defined, but may
+be logically false.  This is implemented by a simple wrapper to the
+check_func for the attribute.
+
+=back
+
+These functions are automatically defined for the built-in Tangram
+types.  In some cases, extra information about what may be put into
+the attribute is cleaned from the SQL column type definition.  This is
+covered by the section on C<parse_X> functions, below.
+
+=head2 Default Type Checking
+
+ # The following list is eval'ed from this documentation, and used as
+ # default attribute options for the specified types.  So, eg, the
+ # default "init_default" for "set" types is a subroutine that returns
+ # a new Set::Object container.
+
+ # "parse" is special - it is passed the options hash given by the
+ # user and should return (\&check_func, \&destroy_func).  This is how
+ # the magical string type checking is performed - see the entry for
+ # parse_string(), below.
+
+ int         => { check_func   => \&check_int },
+ real        => { check_func   => \&check_real },
+ string      => { parse        => \&parse_string },
+ ref         => { check_func   => \&check_obj,
+		  destroy_func => \&destroy_ref },
+ array       => { check_func   => \&check_array,
+		  destroy_func => \&destroy_array },
+ iarray      => { check_func   => \&check_array,
+		  destroy_func => \&destroy_array },
+ flat_array  => { check_func   => \&check_flat_array },
+
+ set         => { check_func   => \&check_set,
+		  destroy_func => \&destroy_set,
+		  init_default => sub { Set::Object->new() }, },
+
+ iset        => { check_func   => \&check_set,
+		  destroy_func => \&destroy_set,
+		  default      => sub { Set::Object->new() }, },
+
+ dmdatetime  => { check_func   => \&check_dmdatetime },
+ rawdatetime => { check_func   => \&check_rawdatetime },
+ rawdate     => { check_func   => \&check_rawdate },
+ rawtime     => { check_func   => \&check_rawtime },
+ flat_hash   => { check_func   => \&check_flat_hash },
+ transient   => { check_func   => \&check_nothing },
+ hash        => { check_func   => \&check_hash,
+		  destroy_func => \&destroy_hash },
+ perl_dump   => { check_func   => \&check_nothing }
+
+=over
+
 =item check_X (\$value)
 
-This series of internal functions checks that $value is of the type X,
-and within applicable bounds.  If there is a problem, then it will
-croak() the error.  These functions are not called from the code, but
-by the set() method on a particular attribute.
-
-When Class::Tangram comes across a type in the schema, it first sets
-up the per-attribute code references to the "check" and "destroy"
-functions.  If there is no check/destroy in the below table, there
-must be a "parse", which is a parse_X function.  See parse_X, below
-
-=head2 TANGRAM TYPE TO FUNCTION MAPPING
-
- int         => { check => \&check_int },
- real        => { check => \&check_real },
- string      => { parse => \&parse_string },
- ref         => { check => \&check_obj,
-		  destroy => \&destroy_ref },
- array       => { check => \&check_array,
-		  destroy => \&destroy_array },
- iarray      => { check => \&check_array,
-		  destroy => \&destroy_array },
- flat_array  => { check => \&check_flat_array },
- set         => { check => \&check_set,
-		  destroy => \&destroy_set },
- iset        => { check => \&check_set,
-		  destroy => \&destroy_set },
- dmdatetime  => { check => \&check_dmdatetime },
- rawdatetime => { check => \&check_rawdatetime },
- rawdate     => { check => \&check_rawdate },
- rawtime     => { check => \&check_rawtime },
- flat_hash   => { check => \&check_flat_hash },
- transient   => { check => \&check_nothing },
- hash        => { check => \&check_hash,
-		  destroy => \&destroy_hash },
- perl_dump   => { check => \&check_nothing }
+This series of internal functions are built-in C<check_func> functions
+defined for all of the standard Tangram attribute types.
 
 =over
 
@@ -480,8 +587,8 @@ sub check_int {
 =item check_real 
 
 checks that the value is a real number, by stringifying it and
-matching it against (m/^-?\d*(\.\d*)?(e-?\d*)?$/).  In the future, a
-test that isn't a bunch of arse will be used.
+matching it against (C<m/^-?\d*(\.\d*)?(e-?\d*)?$/>).  Inefficient?
+Yes.  Patches welcome.
 
 =cut
 
@@ -647,12 +754,12 @@ sub check_hash {
 
 checks whether Australians like sport
 
-=back
-
 =cut
 
 sub check_nothing { }
 
+=back
+
 =item destroy_X ($instance, $attr)
 
 Similar story with the check_X series of functions, these are called
@@ -856,185 +963,11 @@ checks that the value passed is one of the allowed values.
 
 =back
 
-=item import_schema($class)
-
-Parses a tangram object schema, in "\$${class}::schema" to the
-internal representation used to check types values by set().  Called
-automatically on the first get(), set(), or new() for an object of a
-given class.
-
-This function updates Tangram schema option hashes, with the following
-keys:
-
-=over
-
-=item check_func
-
-supply/override the check_X function for this attribute.
-
-=item destroy_func
-
-supply/override the destroy_X function for this attribute
-
 =back
 
-See the SYNOPSIS section for an example of supplying a check_func in
-an object schema.
-
-=cut
-
-sub import_schema($) {
-    my $class = shift;
-
-    # FIXME - expand the run time type information functions, then use
-    # them for this function.
-
-    eval {
-	my ($fields, $bases, $abstract);
-	{
-	    no strict 'refs';
-	    local $^W=0;
-	    eval {
-		$fields = ${"${class}::schema"}->{fields};
-		$bases = ${"${class}::schema"}->{bases};
-		$abstract = ${"${class}::schema"}->{abstract};
-	    };
-	    if ( !$fields and !$bases ) {
-		# hack to pass the "empty inheritance" test
-		my @stack = @{"${class}::ISA"};
-		my %seen = map { $_ => 1 } $class, __PACKAGE__;
-		$bases = [];
-		while ( my $super = pop @stack ) {
-		    if ( defined ${"${super}::schema"} ) {
-			push @$bases, $super;
-		    } else {
-			push @stack, grep { !$seen{$_}++ }
-			    @{"${super}::ISA"};
-		    }
-		}
-		@$bases
-		    or die ("No schema and no Class::Tangram "
-			    ."superclass for $class; define "
-			    ."${class}::schema!");
-	    }
-	}
-
-	my $check_class = { };
-	my $cleaners_class = { };
-	my $init_defaults_class = { };
-	my $types_class = { };
-	my $defs_class = { };
-
-	# if this is an abstract type, do not allow it to be
-	# instantiated
-	if ($abstract) {
-	    $abstract{$class} = 1;
-	}
-
-	# If there are any base classes, import them first so that the
-	# check, cleaners and init_defaults can be inherited
-	if (defined $bases) {
-	    (ref $bases eq "ARRAY")
-		or die "bases not an array ref for $class";
-
-	    # Note that the order of your bases is significant, that
-	    # is if you are using multiple iheritance then the later
-	    # classes override the earlier ones.
-	    for my $super ( @$bases ) {
-		import_schema $super unless (exists $check{$super});
-
-		# copy each of the per-class configuration hashes to
-		# this class as defaults.
-		my ($k, $v);
-
-		# FIXME - this repetition of code is getting silly :)
-		$types_class->{$k} = $v
-		    while (($k, $v) = each %{ $types{$super} } );
-		$check_class->{$k} = $v
-		    while (($k, $v) = each %{ $check{$super} } );
-		$cleaners_class->{$k} = $v
-		    while (($k, $v) = each %{ $cleaners{$super} } );
-		$defs_class->{$k} = $v
-		    while (($k, $v) = each %{ $attribute_options{$super} } );
-		$init_defaults_class->{$k} = $v
-		    while (($k, $v) = each %{ $init_defaults{$super} } );
-	    }
-	}
-
-	# iterate over each of the *types* of fields (string, int, ref, etc.)
-	while (my ($type, $v) = each %$fields) {
-	    if (ref $v eq "ARRAY") {
-		$v = { map { $_, undef } @$v };
-	    }
-	    my $def = $defaults{$type};
-
-	    # iterate each of the *attributes* of a particular type
-	    while (my ($attribute, $option) = each %$v) {
-		$types_class->{$attribute} = $type;
-
-		# ----- check_X functions ----
-		$defs_class->{$attribute} = $option || {};
-
-		if (ref $option eq "HASH" and $option->{check_func}) {
-		    # user-supplied check_X function
-		    $check_class->{$attribute} =
-			$option->{check_func};
-
-		} else {
-		    if (not defined $def) {
-			die "No default check function for type $type";
-		    }
-
-		    # check for a type-specific option hash parser
-		    if ($def->{parse}) {
-			my ($check, $destroy) =
-			    $def->{parse}->($attribute, $option);
-
-			$check_class->{$attribute} = $check;
-			$cleaners_class->{$attribute} = $destroy
-			    if (defined $destroy);
-
-		    } else {
-			# use the default for this type
-			$check_class->{$attribute} = $def->{check};
-		    }
-		}
-
-		# ----- destroy_X functions
-		if (ref $option eq "HASH" and $option->{destroy_func}) {
-		    # user-supplied destroy_X function
-		    $cleaners_class->{$attribute} =
-			$option->{destroy_func};
-		} else {
-		    if ($def->{destroy}) {
-			# use the default for this type
-			$cleaners_class->{$attribute} =
-			    $def->{destroy};
-		    }
-		}
-
-		# ----- init_default functions
-		# create empty Set::Object containers as necessary
-		if ($type =~ m/^i?set$/) {
-		    $init_defaults_class->{$attribute} =
-			sub { Set::Object->new() };
-		}
-		if (ref $option eq "HASH" and $option->{init_default}) {
-		    $init_defaults_class->{$attribute} =
-			$option->{init_default};
-		}
-	    }
-	}
-	$types{$class} = $types_class;
-	$check{$class} = $check_class;
-	$cleaners{$class} = $cleaners_class;
-	$init_defaults{$class} = $init_defaults_class;
-	$attribute_options{$class} = $defs_class;
-    };
-
-    $@ && die "$@ while trying to import schema for $class";
-}
+=head2 Quick Object Dumping and Destruction
 
+=over
 
 =item $instance->quickdump
 
@@ -1131,51 +1064,225 @@ sub clear_refs($) {
 
 =back
 
-=head1 Run-time type information
+=head1 FUNCTIONS
+
+The following functions are not intended to be called as object
+methods.
+
+=head2 Schema Import
+
+=over
+
+=item Class::Tangram::import_schema($class)
+
+Parses a tangram object schema, in C<${"${class}::schema"}> to the
+internal representation used to check types values by C<set()>.
+Called automatically on the first C<get()>, C<set()>, or C<new()> for
+an object of a given class.
+
+=cut
+
+sub import_schema($) {
+    my $class = shift;
+
+    # FIXME - expand the run time type information functions, then use
+    # them for this function.
+
+    eval {
+	my ($fields, $bases, $abstract);
+	{
+
+	    # Here, we go hunting around for their defined schema and
+	    # options
+	    no strict 'refs';
+	    local $^W=0;
+	    eval {
+		$fields = ${"${class}::schema"}->{fields};
+		$bases = ${"${class}::schema"}->{bases};
+		$abstract = ${"${class}::schema"}->{abstract};
+	    };
+	    if ( !$fields and !$bases ) {
+		# hack to pass the "empty inheritance" test
+		my @stack = @{"${class}::ISA"};
+		my %seen = map { $_ => 1 } $class, __PACKAGE__;
+		$bases = [];
+		while ( my $super = pop @stack ) {
+		    if ( defined ${"${super}::schema"} ) {
+			push @$bases, $super;
+		    } else {
+			push @stack, grep { !$seen{$_}++ }
+			    @{"${super}::ISA"};
+		    }
+		}
+		@$bases
+		    or die ("No schema and no Class::Tangram "
+			    ."superclass for $class; define "
+			    ."${class}::schema!");
+	    }
+	}
+
+	# if this is an abstract type, do not allow it to be
+	# instantiated
+	if ($abstract) {
+	    $abstract{$class} = 1;
+	}
+
+	# If there are any base classes, import them first so that the
+	# check, cleaners and init_defaults can be inherited
+	if (defined $bases) {
+	    (ref $bases eq "ARRAY")
+		or die "bases not an array ref for $class";
+
+	    # Note that the order of your bases is significant, that
+	    # is if you are using multiple iheritance then the later
+	    # classes override the earlier ones.
+	    for my $super ( @$bases ) {
+		import_schema $super unless (exists $check{$super});
+
+		# copy each of the per-class configuration hashes to
+		# this class as defaults.
+		my ($k, $v);
+
+		# FIXME - this repetition of code is getting silly :)
+		$types{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $types{$super} } );
+		$check{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $check{$super} } );
+		$cleaners{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $cleaners{$super} } );
+		$attribute_options{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $attribute_options{$super} } );
+		$init_defaults{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $init_defaults{$super} } );
+		$required_attributes{$class}->{$k} = $v
+		    while (($k, $v) = each %{ $required_attributes{$super} } );
+	    }
+	}
+
+	# iterate over each of the *types* of fields (string, int, ref, etc.)
+	while (my ($type, $v) = each %$fields) {
+	    if (ref $v eq "ARRAY") {
+		$v = { map { $_, undef } @$v };
+	    }
+	    my $def = $defaults{$type};
+
+	    # iterate each of the *attributes* of a particular type
+	    while (my ($attribute, $options) = each %$v) {
+
+		# this is what we are finding out about each attribute
+		# $type is already set
+		my ($default, $check_func, $required, $cleaner);
+
+		# set defaults from what they give
+		if (ref $options eq "HASH") {
+		    ($check_func, $default, $required, $cleaner)
+			= @{$options}{qw(check_func init_default
+					 required destroy_func)};
+		}
+
+		# Full their settings with info from defaults
+		if (ref $def eq "HASH") {
+
+		    # try to magically parse their options
+		    if ( $def->{parse} and !($check_func and $cleaner) ) {
+			my @a = $def->{parse}->($attribute, $options);
+			$check_func ||= $a[0];
+			$cleaner ||= $a[1];
+		    }
+
+		    # fall back to defaults for this class
+		    $check_func ||= $def->{check_func};
+		    $cleaner ||= $def->{destroy_func};
+		    $default ||= $def->{init_default};
+		}
+
+		# everything must be checked!
+		die "No check function for ${class}\->$attribute (type $type)"
+		    unless (ref $check_func eq "CODE");
+
+		# for "required" attributes, insert an additional check
+		if ( defined $required ) {
+		    my $old_check_func = $check_func;
+		    my $req = $required;
+		    $check_func = sub {
+			croak "undef not allowed"
+			    unless defined ${$_[0]};
+		        croak "empty value not allowed"
+			  if ($req && ${$_[0]} eq '');
+			goto $old_check_func;
+		    };
+	        }
+
+		$types{$class}->{$attribute} = $type;
+		$check{$class}->{$attribute} = $check_func;
+		$cleaners{$class}->{$attribute} = $cleaner
+		    if (defined $cleaner);
+		$init_defaults{$class}->{$attribute} = $default
+		    if (defined $default);
+		$required_attributes{$class}->{$attribute} = $required
+		    if (defined $required);
+		$attribute_options{$class}->{$attribute} =
+		    ( $options || {} );
+	    }
+	}
+    };
+
+    $@ && die "$@ while trying to import schema for $class";
+}
+
+=back
+
+=head2 Run-time type information
 
 It is possible to access the data structures that Class::Tangram uses
 internally to verify attributes, create objects and so on.
 
-Class::Tangram keeps six internal hashes:
+Class::Tangram keeps seven internal hashes:
 
 =over
 
-=item %types
+=item C<%types>
 
-$types{$class}->{$attribute} is the tangram type of each attribute, ie
-"ref", "iset", etc.  See L<Tangram::Type>.
+C<$types{$class}-E<gt>{$attribute}> is the tangram type of each attribute,
+ie "ref", "iset", etc.  See L<Tangram::Type>.
 
-=item %attribute_options
+=item C<%attribute_options>
 
-$attribute_options{$class}->{$attribute} is the options hash for a
-given attribute.
+C<$attribute_options{$class}-E<gt>{$attribute}> is the options hash
+for a given attribute.
 
-=item %check
+=item C<%required_attributes>
 
-$check{$class}->{$attribute} is a function that will be passed a
-reference to the value to be checked and either throw an exception
+C<$required_attributes{$class}-E<gt>{$attribute}> is the 'required'
+option setting for a given attribute.
+
+=item C<%check>
+
+C<$check{$class}-E<gt>{$attribute}> is a function that will be passed
+a reference to the value to be checked and either throw an exception
 (die) or return true.
 
-=item %cleaners
+=item C<%cleaners>
 
-$attribute_options{$class}->{$attribute} is a reference to a
+C<$attribute_options{$class}-E<gt>{$attribute}> is a reference to a
 destructor function for that attribute.  It is called as an object
 method on the object being destroyed, and should ensure that any
 circular references that this object is involved in get cleared.
 
-=item %abstract
+=item C<%abstract>
 
-$abstract->{$class} is set if the class is abstract
+C<$abstract-E<gt>{$class}> is set if the class is abstract
 
-=item %init_defaults
+=item C<%init_defaults>
 
-$init_defaults{$class}->{$attribute} represents what an attribute is
-set to automatically if it is not specified when an object is
-created. If this is a scalar value, the attribute is set to the
-value. If it is a function, then that function is called (as a method)
-and should return the value to be placed into that attribute.  If it
-is a hash ref or an array ref, then that structure is COPIED in to the
-new object.  If you don't want that, you can do something like this:
+C<$init_defaults{$class}-E<gt>{$attribute}> represents what an
+attribute is set to automatically if it is not specified when an
+object is created. If this is a scalar value, the attribute is set to
+the value. If it is a function, then that function is called (as a
+method) and should return the value to be placed into that attribute.
+If it is a hash ref or an array ref, then that structure is COPIED in
+to the new object.  If you don't want that, you can do something like
+this:
 
    [...]
     flat_hash => {
@@ -1194,7 +1301,7 @@ this information.
 
 =over
 
-=item attribute_options($class)
+=item Class::Tangram::attribute_options($class)
 
 Returns a hash ref to a data structure from attribute names to the
 option hash for that attribute.
@@ -1206,7 +1313,7 @@ sub attribute_options($) {
     return $attribute_options{$class};
 }
 
-=item attribute_types($class)
+=item Class::Tangram::attribute_types($class)
 
 Returns a hash ref from attribute names to the tangram type for that
 attribute.
@@ -1218,7 +1325,20 @@ sub attribute_types($) {
     return $types{$class};
 }
 
-=item known_classes
+=item Class::Tangram::required_attributes($class)
+
+Returns a hash ref from attribute names to the 'required' option setting for
+that attribute.  May also be called as a method, as in
+C<$instance-E<gt>required_attributes>.
+
+=cut
+
+sub required_attributes($) {
+    my $class = ref $_[0] || $_[0];
+    return $required_attributes{$class};
+}
+
+=item Class::Tangram::known_classes
 
 This function returns a list of all the classes that have had their
 object schema imported by Class::Tangram.
@@ -1229,7 +1349,7 @@ sub known_classes {
     return keys %types;
 }
 
-=item is_abstract($class)
+=item Class::Tangram::is_abstract($class)
 
 This function returns true if the supplied class is abstract.
 
@@ -1242,8 +1362,6 @@ sub is_abstract {
     exists $cleaners{$class} or import_schema($class);
 }
 
-=back
-
 =item Class->set_init_default(attribute => $value);
 
 Sets the default value on an attribute for newly created "Class"
@@ -1263,6 +1381,10 @@ sub set_init_default {
     }
 }
 
+=back
+
+=cut
+
 # a little embedded package
 
 package Tangram::Transient;
@@ -1296,22 +1418,27 @@ B<A guided tour of Tangram, by Sound Object Logic.>
 
 =head1 BUGS/TODO
 
-More AUTOLOAD methods, in particular for container types such as
-array, hash, etc.  [on second thoughts, this might not be necessary.
-You can always return a tied hash].
+More AUTOLOAD methods along the line of C<x_includes>, in particular
+for container types such as array, hash, etc.  Either that or
+encourage people to return tied variables if they want to extend that
+particular functionality.
 
 There should be more functions for breaking loops; in particular, a
-standard function called drop_refs($obj), which replaces references to
-$obj with the appropriate Tangram::RefOnDemand so that an object can
-be unloaded via Tangram::Storage->unload() and actually have a hope of
-being reclaimed.  Another function that would be handy would be a
-deep "mark" operation for mark & sweep garbage collection.
+standard function called C<drop_refs($obj)>, which replaces references
+to $obj with the appropriate C<Tangram::RefOnDemand> object so that an
+object can be unloaded via C<Tangram::Storage->unload()> and actually
+have a hope of being reclaimed.  Another function that would be handy
+would be a deep "mark" operation for manual mark & sweep garbage
+collection.
 
-Need to think about writing some functions using Inline for speed.
+Need to think about writing some functions using C<Inline> for speed.
+One of these days...
 
-Allow init_default to be set in import function?
+Allow C<init_default> values to be set in a default import function?
 
-non-persistent attributes
+ie
+
+  use MyClassTangramObject -defaults => { foo => "bar" };
 
 =head1 AUTHOR
 
